using System;
using System.Text.RegularExpressions;

namespace Pentium4emulator
{
	/// <summary>
	/// An enumeration of the different addressing modes supported
	/// Used to mark objects of type "Operand"
	/// </summary>
	public enum addressingMode
	{
		///<summary>
		///No operand
		///</summary>
		nullOperand=0,
		///<summary>
		///An immediate constant e.g. $4
		///</summary>
		immediateValue,
		///<summary>
		///An Immediate register value, e.g. %eax
		///</summary>
		immediateRegister,
		///<summary>
		///Indirect addressing of memory, e.g. (%esp)
		///</summary>
		indirectByRegister,
		///<summary>
		///Immediate offset and base, e.g. $4(%esp)
		///</summary>
		offsetBase,
		///<summary>
		///Two register base and offset, e.g. (%esp, %ecx)
		///</summary>
		baseOffset,
		///<summary>
		///Base, Offset and Scaling factor, e.g. (%esi, %ecx, 4)
		///</summary>
		baseOffsetScaling,
		///<summary>
		///Imm.Offset, Base, Offset register and scaling factor,
		///e.g. 16(%esi, %ecx, 4)
		///</summary>
		offsetBaseOffsetScaling,
		///<summary>
		///to a named variable in RAM
		///</summary>
		namedVariable
	}


	/*
	 * Regular Expressions that match the addressing modes above
	 * 
	 * nullOperand
	 * immediateValue					^\$-?(0x)?[0-9]+
	 * immediateRegister				^\%[a-z]+
	 * indirectByRegister				^\(\s*\%[a-z]+\s*\)
	 * offsetBase						^-?[0-9]+\s*\(\s*\%[a-z]+\s*\)
	 * baseOffset						^\(\s*\%[a-z]+\s*,\s*\%[a-z]+\s*\)
	 * baseOffsetScaling				^\(\s*\%[a-z]+\s*,\s*\%[a-z]+\s*\,\s*[0-9]+\s*\)
	 * offsetBaseOffsetScaling			^[0-9]+\s*\(\s*\%[a-z]+\s*,\s*\%[a-z]+\s*\,\s*[0-9]+\s*\)
	 * namedVariable					^[A-Za-z_]+[A-Za-z_0-9]*
	 * 
	 */


	/// <summary>
	/// Class that holds information about one operand for an opcode.
	/// </summary>
	public class Operand
	{
		#region Regular Expressions declarations.
		private const string kszImmediateValue = "^\\$-?(0x)?[0-9]+";
		private const string kszimmediateregister = "^\\%[a-z]+";
		private const string kszindirectByRegister = "^\\(\\s*\\%[a-z]+\\s*\\)";
		private const string kszoffsetBase = "^-?[0-9]+\\s*\\(\\s*\\%[a-z]+\\s*\\)";
		private const string kszbaseOffset = "^\\(\\s*\\%[a-z]+\\s*,\\s*\\%[a-z]+\\s*\\)";
		private const string kszbaseOffsetScaling = "^\\(\\s*\\%[a-z]+\\s*,\\s*\\%[a-z]+\\s*\\,\\s*[0-9]+\\s*\\)";
		private const string kszoffsetBaseOffsetScaling = "^[0-9]+\\s*\\(\\s*\\%[a-z]+\\s*,\\s*\\%[a-z]+\\s*\\,\\s*[0-9]+\\s*\\)";
		private const string ksznamedVariable = "^[A-Za-z_]+[A-Za-z_0-9]*";

		private static Regex regImmediateValue = new Regex(kszImmediateValue);
		private static Regex regimmediateregister = new Regex(kszimmediateregister);
		private static Regex regindirectByRegister = new Regex(kszindirectByRegister);
		private static Regex regoffsetBase = new Regex(kszoffsetBase);
		private static Regex regbaseOffset = new Regex(kszbaseOffset);
		private static Regex regbaseOffsetScaling = new Regex(kszbaseOffsetScaling);
		private static Regex regoffsetBaseOffsetScaling = new Regex(kszoffsetBaseOffsetScaling);
		private static Regex regnamedVariable = new Regex(ksznamedVariable);
		#endregion

		/// <summary>
		/// Constructor that inits the value by parsing the string
		/// </summary>
		/// <param name="s">The string that can be parsed to get the operand</param>
		public Operand(string s) {
			setValue(s);
		}
		/// <summary>
		/// default Constructor
		/// </summary>
		public Operand() {
		}

		private addressingMode _mode;
		/// <summary>
		/// The addressing mode of this operand. One of the addressingMode enum Values.
		/// </summary>
		public addressingMode AddressingMode {
			get { return _mode; }
			set { _mode = value; }
		}

		/// <summary>
		/// The first element of type 'Register' that makes up the addressing mode of this operand.
		/// This is relevant only if the AddressingMode property is set so that this is used.
		/// </summary>
		public p4engine.regtype reg1;
		/// <summary>
		/// The second element of type 'Register' that makes up the addressing mode of this operand.
		/// This is relevant only if the AddressingMode property is set so that this is used.
		/// </summary>
		public p4engine.regtype reg2;
		/// <summary>
		/// The first element of type 'immediate value' that makes up the addressing mode of this operand.
		/// This is relevant only if the AddressingMode property is set so that this is used.
		/// </summary>
		public int imm1;
		/// <summary>
		/// The second element of type 'immediate value' that makes up the addressing mode of this operand.
		/// This is relevant only if the AddressingMode property is set so that this is used.
		/// </summary>
		public int imm2;
		/// <summary>
		/// The name in case this is a named Variable.
		/// </summary>
		public string name;

		public void setValue(string s) {
			s = s.Trim();
			//Null operand
			if(s == null || s.Length < 1) {
				this.AddressingMode = addressingMode.nullOperand;
			}
			//Immediate Value
			else if(regImmediateValue.IsMatch(s, 0)) {
				this.AddressingMode = addressingMode.immediateValue;
				s = s.Substring(1).Trim();
				if(s.StartsWith("0x"))
					this.imm1 = int.Parse(s.Substring(2), System.Globalization.NumberStyles.HexNumber);
				else 
					this.imm1 = int.Parse(s);
			}
			//Immediate Register
			else if(regimmediateregister.IsMatch(s, 0)) {
				this.AddressingMode = addressingMode.immediateRegister;
				this.reg1 = getRegisterFromString(s.Substring(1));
			}
			//Indirect By register
			else if(regindirectByRegister.IsMatch(s, 0)) {
				this.AddressingMode = addressingMode.indirectByRegister;
				s = s.Substring(1, s.Length - 2).Trim(); // this gets the '%eax' part
				this.reg1 = getRegisterFromString(s.Substring(1));//this gets the getReg('eax')
			}
			//offset-Base
			else if(regoffsetBase.IsMatch(s, 0)) {
				this.AddressingMode = addressingMode.offsetBase;
				//get the immediate value string
				string szImm = s.Substring(0, s.IndexOf('(')).Trim();
				this.imm1 = int.Parse(szImm);
				s = s.Substring(s.IndexOf('(') + 1);// get the '<whitespace>%esi<whitespace>)' part
				s = s.Substring(0, s.Length - 1).Trim();//get the %esi part.
				this.reg1 = getRegisterFromString(s.Substring(1));
			}
			//base-offset
			else if(regbaseOffset.IsMatch(s, 0)) {
				this.AddressingMode = addressingMode.baseOffset;
				//strip off the brackets
				s = s.Substring(1, s.Length - 2).Trim();
				//get the individual regstrings
				string[] sarr = s.Split(new char[] { ',' });
				this.reg1 = getRegisterFromString(sarr[0].Trim().Substring(1));
				this.reg2 = getRegisterFromString(sarr[1].Trim().Substring(1));
			}
			//base-offset-scaling factor
			else if(regbaseOffsetScaling.IsMatch(s, 0)) {
				this.AddressingMode = addressingMode.baseOffsetScaling;
				//strip off the brackets
				s = s.Substring(1, s.Length - 2).Trim();
				//get the individual regstrings
				string[] sarr = s.Split(new char[] { ',' });
				this.reg1 = getRegisterFromString(sarr[0].Trim().Substring(1));
				this.reg2 = getRegisterFromString(sarr[1].Trim().Substring(1));
				this.imm1 = int.Parse(sarr[2].Trim());
			}
			//offset-base-offset-scaling factor
			else if(regoffsetBaseOffsetScaling.IsMatch(s, 0)) {
				this.AddressingMode = addressingMode.offsetBaseOffsetScaling;
				//get the immediate value string
				string szImm = s.Substring(0, s.IndexOf('(')).Trim();
				this.imm2 = int.Parse(szImm);
				s = s.Substring(s.IndexOf('('));//already trimmed
				//strip off the brackets
				s = s.Substring(1, s.Length - 2).Trim();
				//get the individual regstrings
				string[] sarr = s.Split(new char[] { ',' });
				this.reg1 = getRegisterFromString(sarr[0].Trim().Substring(1));
				this.reg2 = getRegisterFromString(sarr[1].Trim().Substring(1));
				this.imm1 = int.Parse(sarr[2].Trim());
			}
			else if(regnamedVariable.IsMatch(s, 0)) {
				this.AddressingMode = addressingMode.namedVariable;
				this.name = s;
			}
			else {
				throw new Exception("Addressing Mode Unrecognized");
			}
		}


		/// <summary>
		/// Parses the input string to a register type.
		/// </summary>
		/// <param name="regname">The string to parse, stripped off the '%'</param>
		/// <returns></returns>
		private p4engine.regtype getRegisterFromString(string regname) {
			if(regname[0] == 'e') {
				if(regname[2] == 'x') {
					switch(regname[1]) {
						case 'a': return p4engine.regtype.eax;
						case 'b': return p4engine.regtype.ebx;
						case 'c': return p4engine.regtype.ecx;
						case 'd': return p4engine.regtype.edx;
						default: throw new Exception("13," + regname);
					}
				}
				else if(regname[2] == 'p') {
					if(regname[1] == 's') return p4engine.regtype.esp;
					else if(regname[1] == 'b') return p4engine.regtype.ebp;
					else throw new Exception("13," + regname);
				}
				else if(regname[2] == 'i') {
					if(regname[1] == 's') return p4engine.regtype.esi;
					else if(regname[1] == 'd') return p4engine.regtype.edi;
					else throw new Exception("13," + regname);
				}
				else throw new Exception("13," + regname);
			}
			else throw new Exception("13," + regname);
		}

	}

}